// Config.cs
// 
// Copyright (C) 2009,2010,2011 Bartek Jasicki
//
// This file is part of Grubng.
//
// Grubng is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

/**@file Config.cs
 * Provide functions for config file.*/

using System;
using System.Xml;

namespace urlsdb
{
	
	/// <summary>
	/// Provide functions for config file.
	/// </summary>
	internal static class Config
	{
		
		/// <summary>
		/// Function create grub.xml file and write default configuration
		/// </summary>
		public static void CreateConfig()
		{
			using (XmlTextWriter xmlw = new XmlTextWriter("grub.xml", null))
			{
				xmlw.Formatting = Formatting.Indented;
				xmlw.WriteStartDocument(false);
				xmlw.WriteStartElement("configuration");
				xmlw.WriteElementString("enablesolr", "Y");
				xmlw.WriteElementString("solraddress", "http://soap.grub.org:8180/solr/");
				xmlw.WriteElementString("solrusername", "user");
				xmlw.WriteElementString("solrpassword", "password");
				xmlw.WriteElementString("uploadaddress", "http://grub.silc.org.ua:8080");
				xmlw.WriteElementString("uploadusername", "user");
				xmlw.WriteElementString("uploadpassword", "password");
				xmlw.WriteElementString("mysqlhost", "localhost");
				xmlw.WriteElementString("mysqldb", "grub");
				xmlw.WriteElementString("mysqluser", "user");
				xmlw.WriteElementString("mysqlpassword", "password");
				xmlw.WriteElementString("workunitsdirectory", "/home/thindil/workunits/wu/");
				xmlw.WriteElementString("workunitspassword", "password");
				xmlw.WriteElementString("useragent", "GrubNG 0.1 (http://grub.org)");
				xmlw.WriteElementString("urlsamount", "250");
				xmlw.WriteElementString("httpversion", "1.0");
				xmlw.WriteElementString("accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
				xmlw.WriteEndElement();
				xmlw.Flush();
				xmlw.Close();	
			}
		}
		
		/// <summary>
		/// Function read data from grub.xml file.
		/// </summary>
		/// <param name="node">
		/// A <see cref="System.String"/> XML node which been read
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/> value of selected XML node
		/// </returns>
		public static string ReadConfig(string node)
		{
			XmlDocument doc = new XmlDocument();
			try
			{
				doc.Load("grub.xml");
			}
			catch (XmlException)
			{
				return "Invalid XML";
			}
			string answer = null;
			try
			{
				XmlNode noder = doc.SelectSingleNode(node);
				answer = noder.InnerText;
				doc = null;
			}
			catch (System.NullReferenceException)
			{
				answer = String.Empty;
			}
			return answer;
		}
		
		/// <summary>
		/// Function insert new data to grub.xml file.
		/// </summary>
		/// <param name="data">
		/// A <see cref="System.String"/> value which been inserted to file
		/// </param>
		/// <param name="name">
		/// A <see cref="System.String"/> name of XML Element which been inserted to file
		/// </param>
		static void InsertConfig(string data, string name)
		{
			XmlDocument doc = new XmlDocument();
			doc.Load("grub.xml");
			XmlNode nodew = doc.SelectSingleNode("/configuration");
			XmlNode childNode = doc.CreateNode(XmlNodeType.Element, name, String.Empty);
			childNode.InnerText = data;
			childNode = nodew.AppendChild(childNode);
			doc.Save("grub.xml");
			doc = null;
		}
		
		/// <summary>
		/// Function update config file.
		/// </summary>
		public static void UpdateConfig()
		{
			XmlDocument doc = new XmlDocument();
			doc.Load("grub.xml");
			string[] nodes = new string[17] {"enablesolr", "solraddress", "solrusername", 
				"solrpassword", "uploadaddress", "uploadusername", "uploadpassword", "mysqlhost", 
				"mysqldb", "mysqluser", "mysqlpassword", "workunitsdirectory", "workunitspassword", 
				"useragent", "urlsamount", "httpversion", "accept"};
			string[] values = new string[17] {"Y", "http://soap.grub.org:8180/solr/", 
				"user", "password", "http://grub.silc.org.ua:8080", "user", "password", "localhost", 
				"grub", "user", "password",	"/home/thindil/workunits/wu/", "password", 
				"GrubNG 0.1 (http://grub.org)", "250", "1.0", 
				"text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"};
			XmlNode noder;
			foreach (string node in nodes)
			{
				noder = doc.SelectSingleNode("/configuration/" + node);
				if (noder == null)
				{
					int index = Array.IndexOf(nodes, node);
					Config.InsertConfig(values[index], node);
				}
			}
			doc = null;
		}
	}
}
